from .architectures import create_ClassificationTransformerModel_v1
from .pipeline_component import (
    make_classification_transformer,
    ClassificationTransformer,
)

from .about import __documentation__, __download_url__, __title__, __version__
